import toml, pathlib, subprocess, sys, pandas
from snakemake import shell

def combine_kraken(inputs):

    tab = pandas.DataFrame()

    for i in inputs:
        tml = open_toml(i)
        isolate = list(tml.keys())[0]
        if tml[isolate]['kraken']['done'] == 'Yes':
            top_3 = tml[isolate]['kraken']['top_3']
            d = {
                'Isolate':isolate,
                'Match #1': top_3[0][0],
                '%1':top_3[0][1],
                'Match #2': top_3[1][0],
                '%2':top_3[1][1],
                'Match #3': top_3[2][0],
                '%3':top_3[2][1],
            }
            df = pandas.DataFrame(d, index = [0])
            df['Quality'] = 'PASS'
        else:
            df = pandas.DataFrame(data = {'Isolate': isolate,
            'Quality': 'kmer-id not determined',
            'Match #1': '', '%1': '', 'Match #2': '', '%2': '', 'Match #3': '', '%3':''},
            index = [0])
        
        if tab.empty:
            tab = df
        else:
            tab = tab.append(df, sort = True)
    tab = tab[['Isolate', 'Match #1', '%1', 'Match #2', '%2', 'Match #3', '%3', 'Quality']]
    tab.to_csv('mtb.tab', sep= '\t', index = False)
    subprocess.run(f"sed -i 's/%[0-9]/%/g' mtb.tab", shell=True)
    return tab

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs):
    
    # set up data dict
    tab = combine_kraken(inputs)
    data = {}
    data['kraken'] = tab.to_dict(orient= 'records')
    write_toml(data = data, output= f'kraken.toml')

inputs = snakemake.input

main(inputs = inputs)


import toml, pathlib, subprocess, sys, pandas

from snakemake import shell

from Bio import SeqIO
                
                
def check_snippy(data, isolate, minaln, aln):
    
    p = pathlib.Path(aln)
    fasta = p.open()
    for i in SeqIO.parse(fasta,'fasta'): # use BioPython to determine percent alignment
        length = len(i.seq)
        nocov = i.seq.count('-')
        lowcov = i.seq.count('N')
        het = i.seq.count('n')
        unaln = nocov + lowcov + het
        perc_aln = 100*(length - unaln) / length
        data[isolate]['qc_snippy']['length'] = length
        data[isolate]['qc_snippy']['nocov'] = nocov
        data[isolate]['qc_snippy']['lowcov'] = lowcov
        data[isolate]['qc_snippy']['het'] = het
        data[isolate]['qc_snippy']['unaln'] = unaln
        data[isolate]['qc_snippy']['pecr_aln'] = perc_aln
        # if the percent alignement is greater than the min alignment
        if perc_aln > float(minaln):
            data[isolate]['qc_snippy']['Quality'] = 'PASS'
        else:
            data[isolate]['qc_snippy']['Quality'] = f'Alignment < {float(minaln)} - isolate will not be included in core'
        return data
                                        

def open_toml(tml):

    data = toml.load(tml)
    print(data)
    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs, isolate, output, minaln):
    

    s = open_toml(tml = inputs)
    print(s)
    rs = s[isolate]['snippy']['run_snippy']
    data = {}
    data[isolate] = {}
    data[isolate]['qc_snippy'] = {}
    data[isolate]['qc_snippy']['run_snippy'] = rs
    if rs == 'Yes':
        data = check_snippy(minaln = minaln,data = data, aln = s[isolate]['snippy']['alignment'], isolate = isolate)
    else:
        data[isolate]['qc_snippy']['Quality'] = 'FAILED sequence QC will not be included in further analysis.'
    
    write_toml(data = data, output = f"{isolate}/snippy_qc.toml") 
    
inputs = snakemake.input
isolate = snakemake.wildcards.sample
output = snakemake.output
minaln = snakemake.params.minaln

main(inputs = inputs, isolate = isolate, output = output,minaln = minaln)

# mash triangle -C *.msh

# mash sketch -m 5 -s 10000 -r -o 2019-12803-6/sketch -I 2019-12803-6 -C 2019-12803-6/R1.fq.gz 2019-12803-6/R1.fq.gz



"""Troika_tb 
.. moduleauthor:: Kristy Horan <kristyhoran15@gmail.com>
"""
__author__ = "Kristy Horan"
__license__ = "GPL"
__maintainer__ = "Kristy Horan"
__email__ = "kristyhoran15@gmail.com"
__status__ = "Development"
__version__ = "0.0.3"
# db_version = "TBProfiler-20190820"


import toml, pathlib, subprocess, sys, pandas, datetime

from snakemake import shell

def get_isolate_list(inputs):
    
    print(inputs)
    isolates = []
    for i in inputs:
        print(i)
        data = open_toml(i)
        print(data)
        isolate = pathlib.Path(i).parts[0]
        if data[isolate]['qc_snippy']['Quality'] == 'PASS' and isolate != "9999-99887":
            isolates.append(isolate)
    
    return ' '.join(isolates)

def core_stats(inputs):
    df = pandas.DataFrame()
    for i in inputs:
        tml = open_toml(i)
        print(tml)
        isolate = list(tml.keys())[0]
        d = pandas.DataFrame(tml[isolate]['qc_snippy'], index = [0])
        d['Isolate'] = isolate
        print(d)
        d = d[['Isolate', 'Quality']]
        if df.empty:
            df = d
        else:
            df = df.append(d)

    # for core.txt
    tab = pandas.read_csv(f"core.txt", sep = '\t')
    print(tab)
    tab['% USED'] = 100 * (tab['LENGTH'] - tab['UNALIGNED'])/ tab['LENGTH']
    tab['% USED'] = tab['% USED'].round(2)
    tab = tab.rename(columns={'ID':'Isolate'})
    tab = pandas.merge(left = df, right = tab, how = 'outer')
    tab = tab[['Isolate','LENGTH', 'ALIGNED','UNALIGNED','VARIANT','HET','MASKED','LOWCOV','% USED', 'Quality']]
    tab.to_csv(f"core_genome.tab", sep='\t', index = False)

    return df.to_dict(orient = 'records')


def generate_snippy_core_cmd(isolates, reference, mask):
    
    cmd = f"snippy-core --mask {mask} --ref {reference} {isolates}" if mask != 'nomask' else f"snippy-core --ref {reference} {isolates}"
    return cmd

def run_cmd(cmd):

    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs, mask, reference):
    # print(inputs)
    # print(mask)
    # print(reference)
    isolates = get_isolate_list(inputs)
    # print(isolates)
    cmd = generate_snippy_core_cmd(isolates = isolates, reference = reference, mask = mask)
    # print(cmd)
    p = run_cmd(cmd)
    data = {}
    data['snippy-core'] = {}
    if p == 0:
        
        date = datetime.datetime.today().strftime("%d_%m_%y")
        data['snippy-core'][date] = {}
        data['snippy-core'][date]['isolates'] = isolates.split()
        data['snippy-core'][date]['reference'] = reference
        data['snippy-core'][date]['mask'] = mask 
        data['snippy-core'][date]['vcf'] = 'core.vcf'
        data['snippy-core'][date]['aln'] = 'core.aln'
        data['snippy-core'][date]['txt'] = 'core.txt'
        data['snippy-core'][date]['tab'] = 'core.tab'
        data['snippy-core'][date]['fullaln'] = 'core.full.aln'
        data['snippy-core'][date]['stats'] = core_stats(inputs)

        write_toml(data = data, output = "snippy_core.toml")
   
inputs = snakemake.input
mask = snakemake.params.mask_string
reference = snakemake.params.reference

main(inputs = inputs, mask = mask, reference = reference)

# mash triangle -C *.msh

# mash sketch -m 5 -s 10000 -r -o 2019-12803-6/sketch -I 2019-12803-6 -C 2019-12803-6/R1.fq.gz 2019-12803-6/R1.fq.gz


#!/usr/bin/env python3.7

import subprocess, pathlib, json, sys, toml

from snakemake import shell


def run_snpit(vcf):
    # print(vcf)   
    v = pathlib.Path(vcf)
    if v.exists():
        cmd = f"snpit --input {v}"
        p = subprocess.run(cmd, shell = True, capture_output = True, encoding = "utf-8")
        snpit_data = p.stdout.split("\n")
        lineage_strs = snpit_data[1].split('\t')
        sp = lineage_strs[1]
        l = lineage_strs[2]
        f = lineage_strs[4] if lineage_strs[4] != 'N/A' else ''
        
        return [sp, l,f]

def append_to_json(lineage, json_file, outfile):
    isolate = pathlib.Path(json_file).name.split('.')[0]
    j = json.load(open(json_file))
    # print(j)
    # d = j
    j['snpit'] = {
        'species':lineage[0],
        'lineage':lineage[1],
        'family':lineage[2]
    }
    
    with open(outfile, 'w') as f:
        json.dump(j, f)
    
    return j

def run_cmd(cmd):

    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    # print(data)
    with open(output, 'w') as f:
        
        toml.dump(data, f)
    
def main(tbprofiler,isolate):
    
    tml = open_toml(tbprofiler)
    # print(tml)
    data = {}
    data[isolate] = {}
    data[isolate]['snpit'] = {}
    data[isolate]['tb_profiler'] = {}
    if tml[isolate]['tb_profiler']['done'] == 'Yes':
        vcf = f"{isolate}/snps.raw.vcf"
        snpit = run_snpit(vcf = vcf)
        # print(snpit[0], snpit[1],snpit[1])
        data[isolate]['tb_profiler']['done'] = 'Yes'
        data[isolate]['tb_profiler']['kmer-id'] = tml[isolate]['tb_profiler']['kmer-id']
        # print(data[isolate])
        data[isolate]['snpit']['species'] = snpit[0]
        # # print(data[isolate]['snpit'])
        data[isolate]['snpit']['lineage'] = snpit[1]
        # # print(data[isolate]['snpit'])
        data[isolate]['snpit']['family'] = snpit[2]
        # print(data[isolate]['snpit'])
        data[isolate]['snpit']['done'] = 'Yes'
        # print(data[isolate]['snpit'])
        append_to_json(lineage = snpit, json_file = f'{isolate}/tbprofiler.results.json', outfile = f'{isolate}/tbprofiler.snpit_results.json')
        # print(isolate)
        
    else:
        data[isolate]['tb_profiler']['done'] = 'No'
        data[isolate]['snpit']['done'] = 'No'
        data[isolate]['tb_profiler']['kmer-id'] = tml[isolate]['tbprofiler']['kmer-id']
        data[isolate]['snpit']['data'] = {}
    write_toml(data = data, output = f"{isolate}/snpit.toml")

    
tbprofiler = snakemake.input.tbprofiler
isolate = snakemake.wildcards.sample

main(tbprofiler = tbprofiler, isolate = isolate)


"""troika
.. moduleauthor:: Kristy Horan <kristyhoran15@gmail.com>


"""


import logging
import argparse
import configargparse
import pathlib
import sys
import os
import datetime
from troika_tb.RunTroika import Troika
from troika_tb.versions import db_version, troika_version


def run_pipeline(args):
    '''
    Run the pipeline
    '''
    T = Troika(args)
    return(T.run_pipeline())

def main():
    # setup the parser
  
    parser = configargparse.ArgumentParser(description='Troika - a pipeline for phylogenentic analysis, detection and reporting of genomic AST in Mtb',formatter_class=configargparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + troika_version)
    # use add_argument
    parser.add_argument('--input_file',
        '-i', 
        help='Input file tab-delimited file 3 columns isolate_id path_to_r1 path_to_r2',
        default = ''
    )
    parser.add_argument('--detect_species',
        '-d',
        action = 'store_true',
        help = 'Set if you would like to detect species - note if not set troika may include non-tuberculosis species in the analysis.'
    )
    parser.add_argument('--resistance_only',
        action = 'store_true',
        help = 'If detection of resistance mutations only is needed. Phylogeny will not be performed.'
    )
    parser.add_argument(
        "--Singularity",
        "-S",
        action="store_true",
        help="If singularity is to be used for troika."
    )
    parser.add_argument(
        '--profiler_singularity_path',
        '-ps',
        help='URL for TB-profiler singularity container.',
        default = 'docker://mduphl/mtbtools'
    )
    parser.add_argument(
        '--snippy_singularity_path',
        '-ss',
        help='URL for Snippy singularity container.',
        default = 'docker://mduphl/snippy:v4.4.3'

    )
    parser.add_argument(
        "--workdir",
        "-w",
        default=f"{pathlib.Path.cwd().absolute()}",
        help="Working directory, default is current directory",
    )
    parser.add_argument(
        "--resources",
        "-r",
        default=f"{pathlib.Path(__file__).parent }",
        help="Directory where templates are stored",
    )
    parser.add_argument(
        "--jobs", "-j", default=8, help="Number of jobs to run in parallel."
    )
    parser.add_argument(
        '--profiler_threads',
        '-t',
        help='Number of threads to run TB-profiler', 
        default=1
    )
    parser.add_argument(
        '--kraken_threads',
        '-kt',
        help = 'Number of threads for kraken',
        default = 4
    )
    parser.add_argument(
        '--kraken_db', 
        '-k', 
        env_var='KRAKEN2_DEFAULT_DB', 
        help='Path to DB for use with kraken2, if no DB present speciation will not be performed.'
    )
    parser.add_argument(
        '--snippy_threads',
        '-st',
        help = 'Number of threads for snippy',
        default = 8
    )
    parser.add_argument(
        '--mode',
        '-m',
        help = "If running for MDU service use 'mdu', else use 'normal'",
        default = 'normal',
        choices = ['mdu', 'normal']
    )
    parser.add_argument(
        '--positive_control',
        '-pc',
        help = 'Path to positive control - REQUIRED if running for MDU service',
        default= '')

    parser.add_argument(
        '--db_version',
        help=f'The version of database being used.', 
        default=db_version
    )
    parser.add_argument(
        '--min_cov',
        '-mc',
        default=40,
        help = f"Minimum coverage for quality checks, isolates with coverage below this threshold will not be used in the analysis."
    )
    parser.add_argument(
        '--min_aln',
        '-ma    ',
        default=80,
        help = f"Minimum alignment for phylogenetic analysis, alignments lower than this threshold will not be included in the calculation of core-genome."
    )
    parser.set_defaults(func=run_pipeline)
    args = parser.parse_args()
    
    if vars(args) == {}:
        parser.print_help(sys.stderr)
    else:
        
        args.func(args)
	
if __name__ == '__main__':
    main()




import toml, pathlib, subprocess, sys, pandas

from snakemake import shell

def get_top_3(isolate):
    
    report = pathlib.Path(isolate, 'kraken2.tab')
    df= pandas.read_csv(report, sep = "\t", header =None, names = ['percentage', 'frag1', 'frag2','code','taxon','name'])
    df['percentage'] = df['percentage'].apply(lambda x:float(x.strip('%')) if isinstance(x, str) == True else float(x)) #remove % from columns
    df = df.sort_values(by = ['percentage'], ascending = False)
    df = df[df['code'].isin(['U','S'])]     
    df = df.reset_index(drop = True) 
    top_three = [
        (f"{df.loc[0,'name'].strip()}",round(df.loc[0,'percentage'],2)),
        (f"{df.loc[1,'name'].strip()}",round(df.loc[1,'percentage'],2)),
        (f"{df.loc[2,'name'].strip()}",round(df.loc[2,'percentage'],2))
    ]

    return top_three

def generate_cmd(r1, r2, isolate, db, data):
    
    
    
    cmd = f"kraken2 --paired {r1} {r2} --minimum-base-quality 13 --report {isolate}/kraken2.tab --memory-mapping --db {db} --threads 4"
    data[isolate]['kraken']['kraken_db'] = f"{db}"

    return cmd, data

def run_cmd(cmd):
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def extract_metrics(mash_string):
    
    for m in mash_string:
        if 'Estimated coverage' in m:
            d = m.split(':')[-1].strip()
    return d
    

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(r1, r2, isolate, kraken_db,run_kraken):
    
    # set up data dict
    tml = f"{pathlib.Path(isolate, 'kraken.toml')}"
    data = {}
    data[isolate] = {}
    data[isolate]['kraken'] = {}
    # run kraken
    if run_kraken:
        cmd, data = generate_cmd(r1 = r1, r2 = r2, isolate = isolate, db = kraken_db, data = data)
        kraken_returncode = run_cmd(cmd)
        if kraken_returncode == 0:
            # add to data dict
            data[isolate]['kraken']['done'] = 'Yes'
            data[isolate]['kraken']['file'] = f"{isolate}/kraken2.tab"
            data[isolate]['kraken']['top_3'] = get_top_3(isolate)
    else:
        data[isolate]['kraken']['done'] = 'No'
    write_toml(data = data, output= f"{isolate}/kraken.toml")

r1 = snakemake.input.r1
r2 = snakemake.input.r2
isolate = snakemake.wildcards.sample
run_kraken = snakemake.params.run_kraken
kraken_db = snakemake.params.kraken_db

main(r1 = r1, r2 = r2, isolate = isolate,run_kraken = run_kraken, kraken_db = kraken_db)


import toml, pathlib, subprocess, sys
from snakemake import shell


def generate_snippy_cmd(r1, r2, isolate, reference, threads):
    
    cmd = f"snippy --outdir {isolate} --ref {reference} --R1 {r1} --R2 {r2} --force --cpus {threads}"
    print(cmd)
    return cmd

def run_cmd(cmd):
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def get_species(kraken):
    mtbc = ['pinnipedii', 'tuberculosis', 'orygis', 'bovis', 'bovis BCG', 'canetti', 'microti', 'africanum']
    tml = open_toml(kraken)
    print(tml)
    isolate = list(tml.keys())[0]
    if tml[isolate]['kraken']['done'] == 'No':
        return 'kmer-id not performed'
    elif tml[isolate]['kraken']['done'] == 'Yes':
        s = tml[isolate]['kraken']['top_3'][0][0]
        # print(s)
        if 'Mycobacterium' in s:
            print('Mycobacterium genus found')
            for m in mtbc:
                print(m)
                print(m in s)
                if m in s:
                    print("found MTBC species")
                    return 'MTBC'

def get_reads(seqdata, isolate):

    s = open_toml(seqdata)
    r1 = s[isolate]['seqdata']['R1']
    r2 = s[isolate]['seqdata']['R2']
    return r1,r2

def get_quality(seqdata, kraken, isolate):

    s = open_toml(seqdata)
    print(s)
    k = get_species(kraken)
    print(k)
    if s[isolate]['seqdata']['data']['Quality'] == 'PASS' and k == 'MTBC':
        return 'Yes'
    else:
        return 'No'


def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(seqdata,kraken, isolate, output, reference, threads):
    
    print(seqdata)
    # print(threads)
    r1,r2 = get_reads(seqdata = seqdata, isolate = isolate)
    run_snippy = get_quality(seqdata = seqdata, kraken = kraken, isolate = isolate)
    print(run_snippy)
    data = {}
    data[isolate] = {}
    data[isolate]['snippy'] = {}
    data[isolate]['snippy']['reference'] = reference
    data[isolate]['snippy']['R1'] = r1
    data[isolate]['snippy']['R2'] = r2
    data[isolate]['snippy']['run_snippy'] = run_snippy

    if run_snippy == 'Yes':
        cmd = generate_snippy_cmd(r1 = r1, r2=r2, isolate = isolate, reference = reference, threads = threads)
        # print(cmd)
        p = run_cmd(cmd)
        # print(p)
        if p == 0:
            data[isolate]['snippy']['bam'] = f"{isolate}/snps.bam"
            data[isolate]['snippy']['alignment'] = f"{isolate}/snps.aligned.fa"
            data[isolate]['snippy']['vcf'] = f"{isolate}/snps.vcf"
            data[isolate]['snippy']['cmd'] = cmd
    write_toml(data = data, output = f"{isolate}/snippy.toml") 

seqdata = snakemake.input.seqdata
kraken = snakemake.input.kraken
isolate = snakemake.wildcards.sample
output = snakemake.output
reference = snakemake.params.reference
threads = snakemake.threads

main(seqdata = seqdata, kraken =kraken, isolate = isolate, output = output,reference = reference, threads =threads)



import toml, pathlib, subprocess, sys, re

from snakemake import shell

def generate_seqdata_cmd(r1, r2, isolate):
    
    cmd = f"seqtk fqchk {r1} {r2} > {isolate}/seqdata.tab"
    return cmd

def run_cmd(cmd):
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode


def get_data(cmd_output):
#    print(seqtkdata[0]
    output = pathlib.Path(f"{cmd_output}").open().readlines()
    summary_pat = re.compile(r'(\w+):\s(\d+\.?\d?\d?);')
    distinct_values = re.compile(r';\s(\d+)\s\w+')
    # print(distinct_values)
    print(output[1])
    record = re.compile(r'(\d+\.?\d?\d?)')
    summary_dict = dict((k, float(v)) for k, v in summary_pat.findall(output[0]))
    # print(distinct_values.findall(output[0]))
    summary_dict['distinct_error_codes'] = int(distinct_values.findall(output[0])[0])
    header = output[1].strip().replace('%', '').replace('#', '').replace("POS\t", "").split("\t")
    line_one = [float(v) for v in record.findall(output[2])]
    summary_dict.update(dict(zip(header, line_one)))
    header.insert(0, 'position')
    summary_dict['Reads'] = dict(zip(header, [float(v) for v in record.findall(output[3])]))['bases']
    summary_dict['GC content'] = round((summary_dict['C'] + summary_dict['G']),2)
    median_position = (summary_dict['Reads'] + 1) / 2
    for line in output[4:]:
        parsed_record = dict(zip(header, [float(v) for v in record.findall(line)]))
        if parsed_record['bases'] > median_position:
            continue
        else:
            summary_dict['med_len'] = parsed_record['position'] - 1
            break
    return(summary_dict)

def get_coverage(mash, isolate):

    m = open_toml(mash)
    print(m)
    return m[isolate]['mash']['Estimated coverage']


def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(r1, r2, isolate, mash, mincov):
    
    msh = open_toml(mash)
    print(msh)
    data = {}
    data[isolate] = {}
    data[isolate]['seqdata'] = {}
    data[isolate]['seqdata']['R1'] = r1
    data[isolate]['seqdata']['R2'] = r2
    data[isolate]['seqdata']['file'] = f"{pathlib.Path(isolate,'seqdata.tab')}"


    cmd = generate_seqdata_cmd(r1 = r1, r2 = r2, isolate = isolate)
    p = run_cmd(cmd)
    # print(p)
    if p == 0:
        cov = get_coverage(mash = mash, isolate=isolate)
        # print(cov)
        data[isolate]['seqdata']['data'] = get_data(cmd_output = data[isolate]['seqdata']['file'])
        data[isolate]['seqdata']['data']['Estimated_coverage'] = round(float(cov),2)
        data[isolate]['seqdata']['data']['Quality'] = 'PASS' if data[isolate]['seqdata']['data']['Estimated_coverage'] > int(mincov) else 'FAIL - will be removed from further analysis'
        write_toml(data = data, output = f"{isolate}/seqdata.toml")   
  
r1 = snakemake.input.r1
r2 = snakemake.input.r2
mash = snakemake.input.mash
isolate = snakemake.wildcards.sample
mincov = snakemake.params.mincov

main(r1 = r1, r2 = r2, isolate = isolate, mash = mash, mincov = mincov)





import svgwrite, re, sys, subprocess, toml
from Bio import Phylo
import jinja2, pathlib, pandas, numpy, re
from packaging import version
import datetime

from snakemake import shell

class Report:
    '''
    A class to generate the tables and figures for use in the report.html
    '''
    def write_tables(self,table):
        '''
        Write a table, given a tab delimited file generate a html string
        '''
        # TODO add class isolate id to <tr>
        # TODO add class distances-isolate to tr if matrix and head-isolate to head td
        path = pathlib.Path(f"{table}")
        if path.exists():
            data = open(path).readlines()
            # for header
            header = data[0].split('\t') 
            if 'distances.tab' in table:
                tablehead = [f"<th class='{column}-head'>{column}</th>" for column in header]
            else:
                tablehead = [f"<th>{column}</th>" for column in header]
            # for body seqtablebody
            body = []
            for i in range(1,len(data)):
                raw = data[i].split('\t')  
                if 'summary_table.tab' in table:
                    row = [f"<tr class='{raw[0]} tiplab'>"]
                elif 'distances.tab' in table:
                    row = [f"<tr class='distances-{raw[0]}'>"]
                elif 'mtb' in table:
                    row = [f"<tr class='{raw[0]}-species-identification'>"]
                elif 'core_genome.tab' in table:
                    row = [f"<tr class='{raw[0]}-core-genome'>"]
                elif 'troika.tab' in table:
                    row = [f"<tr class='{raw[0]}-drug-resistance'>"]
                elif 'seqdata.tab' in table:
                    row = [f"<tr class='{raw[0]}-sequence-data'>"]
                else:
                    row = [f"<tr>"] # TODO add class isolate id to <tr>
                if 'distances.tab' in table:
                    for d in range(len(raw)):
                        row.append(f"<td align=\"center\" class = \"{raw[0]}_{header[d]}\">{raw[d]}</td>")    
                else:
                    for d in raw:
                        row.append(f"<td align=\"center\">{d}</td>")
                row.append(f"</tr>")
                body = body + row
            return('\n'.join(tablehead),'\n'.join(body))

    

    def get_table_data(self,td):
        '''
        input:
            :reportdir: the directory where report files are kept
            :td: the dictionary for data
        output:
            :td: an updated dictionary
        '''

        for tabletype in range(len(td)):
            table = td[tabletype]['file']
            td[tabletype]['head'], td[tabletype]['body'] = self.write_tables(table=table)
        
        return(td)

    def plot_histogram(self,series, xlabel, ylabel, bins, color = "#3973ac", width = 1000, plot_height = 200):
        '''
        generic method for generating a histogram from a dataframe using bokeh
        input:
            :series:the column of the dataframe to be used
            :xlabel: label for x-axis
            :ylabel: label for y-axis
            :bins: the size of the histogram bins
            :color: color of the bars
            :width: width of the graph
            :height: height of the graph
        output:
            :script: the javascript string for insert into html
            :div: the html div for inster in html doc
        '''
        # generate histogram
        hist, edges = numpy.histogram(series, density=True, bins=bins)
        # generate bokeh plot
        p = figure(plot_width=width, plot_height=plot_height)
        p.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:], line_color=color, color = color)
        # style
        p.xaxis.axis_label = xlabel
        p.yaxis.axis_label = ylabel
        p.sizing_mode = "scale_width"
        # save as png
        # export_png(p, filename="snpdensity.png")
        # get div and script and and add to dict
        script, div = components(p)
        # show(p)
        return(script,div)

    def plot_snpdensity(self,idx_file):

        '''
        generate a snp-density accross the genome plot - using the core.tab file
        input:
            :reportdir: the directory where files are kept
        out:
            :snpdensityscript: the javascript string for insert into html
            :snpdenstiydiv: the html div for inster in html doc
        '''
        # helper functions for getting the data into the right format.
        def adjust_offset(row, d):
            if row['CHR'] in d:
                return(int(d[row['CHR']]) + int(row['POS']))
            else:
                return(int(row['POS']))

        def generate_dict(idx_file):
            d = {}
            pos = 0
            offset = 0
            with open(f"{idx_file}") as f:
                for line in f:
                    l = line.split()
                    d[l[0]] = offset
                    offset = offset + int(l[1])
            return(d)
        
        # open fai file and generate the dictionary
        
        d = generate_dict(idx_file)
        # get the core file
        core = 'core.tab'
        df = pandas.read_csv(core, sep = '\t')
        # get a list of isolate names
        names = list(df.columns[3:len(df.columns)])
        # if the there is no snp in the isolate (ie same as ref then mak na - then easy to drop)
        for i in names:
                df[i]=numpy.where(df['REF'] == df[i], numpy.nan, df[i])
        # generate the offset value
        df['POS_OFFSET'] = df[['CHR', 'POS']].apply(lambda x:adjust_offset(x,d), axis = 1)
        # collect positions to get allow for histogram and dropna (no snp)
        melted_df = pandas.melt(df, id_vars=['POS_OFFSET'], value_vars=names)
        melted_df = melted_df.dropna()
        melted_df = melted_df.sort_values(by= ['POS_OFFSET'])
        # generate histogram
        # snpdensityscript, snpdensitydiv = self.plot_histogram(series=melted_df['POS']/1000,xlabel="Genome Position (MB)", ylabel="SNPS",bins=10000)
        contig_breaks = [d[value] for value in d]
        
        # return dictionary
        return(list(melted_df['POS_OFFSET']/1000))

    def plot_distances(self):

        '''
        generate a snp-density plot - using the distacnes.tab file
        input:
            :reportdir: the directory where files are kept
        out:
            :distancescript: the javascript string for insert into html
            :distancesdiv: the html div for inster in html doc
        '''
        distance = 'distances.tab'

        df = pandas.read_csv(distance, sep = '\t')
        # get a list of isolate names
        names = list(df.columns[1:len(df.columns)])
        col1 = df.columns[0]
        
        # if the there is no snp in the isolate (ie same as ref then mak na - then easy to drop)
        
        # collect positions to get allow for histogram and dropna (no snp)
        melted_df = pandas.melt(df, id_vars=[col1], value_vars=names)
        melted_df = melted_df[melted_df[col1]!= melted_df['variable']]
        # generate histogram
        # distancescript, distancediv = self.plot_histogram(series=melted_df['value'], xlabel="SNP distances", ylabel="Frequency", bins=100)
        # td['pairwisedistance'] = {'script':distancecript, 'div':distancediv}
        # return dictionary
        return(list(melted_df['value']))


    def get_tree_string(self):
        '''
        Generate a tree image from a newick
        input:
            :reportdir: the directory where report files are stored
        output:
            string reporesentation of the path to the tree image
        '''
        # get tree

        with open(f"core.treefile", 'r') as t:
            tree = t.read().strip()

        return tree

    def get_software_versions(self, software):

        '''
        Given the name of the software, find the version
        input:
            :software: the name of the software
        output:
            a string in the form of 'Name_of_Sofware v.X.Y.Z'
        '''

        version_pat = re.compile(r'\bv?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<release>[0-9]+)(?:\.(?P<build>[0-9]+))?\b')

        if software == 'snp-dists':
            vs = '-v'
        else:
            vs = '--version'
        cmd = f"{software} {vs} 2>&1"
        
        p = subprocess.run(cmd, shell = True, capture_output=True, encoding = "utf-8")
        sft = p.stdout
        
        v = version_pat.search(sft)
        v = v.group()
        sft_version = f"{software} v.{v}"
        return(sft_version)
    
    def make_dict_versions(self, tools):
        '''
        Called by get_software_file to make a dictionary of tools used
        input:
            :tools: a list of tools
        output:
            a dictionary with tools and versions.
        '''
        tool_dict = {}
        for t in tools:
               v = self.get_software_versions(t)
               tool_dict[t] = v
        return(tool_dict)

    def get_software_file(self):
        '''
        get the versions of software on the system at completion of the piepline
        input:
            :reportdir: the directory where report files are stored
            :pipeline: the type of pipeline
            :assembler: the assembler used in the pipeline
        '''
        tools = ['snippy', 'snippy-core', 'snp-dists', 'iqtree', 'tb-profiler']
        
        tool_dict = self.make_dict_versions(tools)
        
        versions = ['Software versions']
        for t in tool_dict:
            versions.append(tool_dict[t])
        
        p = pathlib.Path('software_versions.tab')

        p.write_text('\n'.join(versions))

    def merge_dfs(self,start, added):
        
        if start.empty:
            start = added
        else:
            start = start.merge(added)
        return(start)

    def generate_summary(self):
        '''
        function to generate a summary table
        '''
        tabs = ['troika.tab', 'mtb.tab', 'distances.tab', 'seqdata.tab','core_genome.tab','core.treefile','core.tab']
        summary_df = pandas.DataFrame()
        df_list = []
        print(tabs)
        for tab in tabs:
            # print(df)
            if tab == 'mtb.tab' and pathlib.Path('mtb.tab').exists():
                species = pandas.read_csv(tab, sep = '\t')
                print(species)
                species = species[['Isolate', 'Match #1']]
                summary_df = self.merge_dfs(summary_df, species)
                summary_df = summary_df.rename(columns={'#1 Match': 'Species'})
            elif 'seqdata' in tab:
                seq = pandas.read_csv(pathlib.Path(tab), sep = '\t')
                seq = seq[['Isolate', 'Estimated depth']]
                summary_df = self.merge_dfs(summary_df, seq)
            elif 'core_genome' in tab:
                core = pandas.read_csv(pathlib.Path(tab), sep = '\t')
                core = core[['Isolate', '% USED']]
                summary_df = self.merge_dfs(summary_df,core)
            elif 'troika' in tab:
                res = pandas.read_csv(pathlib.Path(tab), sep = '\t')
                # print(res.columns)
                res = res[['Isolate', 'Predicted drug resistance', 'Phylogenetic lineage']]
                # res = res.rename(columns = {'MDU ID': 'Isolate'})
                summary_df = self.merge_dfs(summary_df,res)
        if 'Species' not in list(summary_df.columns):
            summary_df['Species'] = 'Species detection not performed'
        summary_file = 'summary_table.tab'
        summary_df.to_csv(summary_file, sep = '\t', index = False)

        return len(summary_df['Isolate'])

    def get_metadata(self):

        resdata = pandas.read_csv('troika.tab', sep = '\t') 
        data = {}
        for row in resdata.iterrows():
            print(row[1])
            rif =  "#adf0ad" if row[1]["Rifampicin"]== "No mutation detected" else "#f03329"
            inh = "#adf0ad" if row[1]["Isoniazid"]== "No mutation detected" else "#f03329"
            pza = "#adf0ad" if row[1]["Pyrazinamide"]== "No mutation detected" else "#f03329"
            emb = "#adf0ad" if row[1]["Ethambutol"]== "No mutation detected" else "#f03329"
            data[row[1]["Isolate"]] = {
                'Rif':{'colour':rif},
                'Inh':{'colour':inh},
                'Pza':{'colour':pza},
                'Emb':{'colour':emb}
            }
        return data,len(list(resdata['Isolate']))

    def open_toml(self,tml):

        data = toml.load(tml)

        return data
    
    def write_toml(self,data, output):
        
        with open(output, 'wt') as f:
            toml.dump(data, f)

    def main(self,workdir, resources, amr_only, idx):
        '''
        main function of the report class ties it all together
        input:
            :workdir: job directory
            :resources: the directory where templates are stored
            :amr_only: whether or not amr_only is being run
        '''
        
        
        # set up paths variables
        p = pathlib.Path('.')
        date = datetime.datetime.today().strftime("%d/%m/%y")
        # path to report data
        reportdir = pathlib.Path(workdir)
        reporthtml = pathlib.Path('index.html')
        # path to html template
        indexhtml = pathlib.Path(resources,'index.html') # replace with template 
        # newick string
        newick_path = 'core.treefile'
        newick_string = open(newick_path).read().strip()
        # save tool table
        self.get_software_file()
        
        # table dictionary for each option
        
        td = [{'file':'seqdata.tab', 'title':'Sequence Data', 'link': 'sequence-data', 'type' : 'table'}, {'file':'summary_table.tab','title':'Summary', 'link':'summary', 'type':'summary'}, {'file':'troika.tab', 'title':'Drug Resistance', 'type':'table', 'link':'drug-resistance'}]
        sp = 'mtb.tab'
        
        # TODO edit links to be title lower case separated by a -
        core_genome_td = {'file': 'core_genome.tab', 'title': 'Core Genome', 'link':'core-genome', 'type':'table'}
        snp_density_td = {'title': 'SNP density', 'link':'snp-density', 'type':'graph'}
        core_phylogeny_td = {'title': 'Phylogeny', 'link':'phylogeny', 'file': 'core.treefile', 'type': 'tree'}
        snp_distance_td = {'file': 'distances.tab', 'title':'SNP distances', 'type':'matrix', 'link':'snp-distances'}
        # list of snp tasks
        s_td = [core_genome_td,snp_density_td,core_phylogeny_td, snp_distance_td]
        # species_id_td = 
        
        # list of assembly tasks       
    
        td.extend(s_td)
        tables =['drug-resistance', 'core-genome', 'snp-distances', 'sequence-data']
        modaltables =['drug-resistance','sequence-data', 'core-genome']
        display = f"display:inline;"
    
        tables.append('versions')
        # get versions of software
        versions_td = {'file': 'software_versions.tab', 'title': 'Tools', 'type': 'versions', 'link':'versions'}
        td.append(versions_td)
        snpdistances = ''
        snpdensity = ''
        # add data to sections
        # print(td)
        for t in range(len(td)):
            # print(t)
            # TODO if table add a modal modal + link and link will be title lowercase with hyphen
            if td[t]['type'] == 'table':
                td[t]['head'], td[t]['body'] = self.write_tables(table=td[t]['file'])
            if td[t]['type'] == 'tree':
                td[t]['image'] = self.get_tree_string()
            if td[t]['link'] == 'snp-distances':
                td[t]['head'], td[t]['body'] = self.write_tables(table=td[t]['file'])
                snpdistances = self.plot_distances()
            if td[t]['link'] == 'snp-density':
                snpdensity= self.plot_snpdensity(idx_file = idx)
            if td[t]['type'] == 'versions':
                td[t]['head'], td[t]['body'] = self.write_tables(table=td[t]['file'])
            if td[t]['type'] == 'summary':
                self.generate_summary()
                td[t]['head'], td[t]['body'] = self.write_tables(table = td[t]['file'])
        metadata_string,height  = self.get_metadata()
        data = {'newick': newick_string, 
                'display': display,
                'tables': tables,
                'td': td, 
                'snpdistances':snpdistances, 
                'snpdensity': snpdensity, 
                'modaltables' : modaltables, 
                'date': date, 
                'metadata_string': metadata_string, 
                'height': height*50}

        self.write_toml(data, 'report.toml')
        # fill template
        
        report_template = jinja2.Template(pathlib.Path(indexhtml).read_text())
        reporthtml.write_text(report_template.render(data))

resources = snakemake.params.template_path
workdir = snakemake.params.wkdir
amr_only = snakemake.params.amr_only
idx = snakemake.params.idx

R = Report()
R.main(resources = resources, workdir = workdir, amr_only = amr_only, idx = idx)


import pathlib
import os, getpass, shutil, re, psutil
import pandas
import jinja2
import sh
import logging
import filecmp
import datetime
import numpy
import itertools
import subprocess
import json
from Bio import SeqIO, Phylo
from packaging import version
from troika_tb.versions import db_version



class Troika(object):
    '''
    A class for Troika
    '''
    
    def __init__(self, args):
        # from logger import logger
        # get date and time
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        # create file handler which logs even debug messages
        fh = logging.FileHandler('troika.log')
        fh.setLevel(logging.INFO)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('[%(levelname)s:%(asctime)s] %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        self.logger.addHandler(fh)
        self.logger.info(f"Initialising Troika")
        self.now = datetime.datetime.today().strftime("%d_%m_%y_%H")
        self.date = datetime.datetime.today().strftime("%d_%m_%y")
        self.input_file = self.check_input_file(args.input_file)
        self.workdir = self.check_input_file(args.workdir)
        self.resources = self.check_input_file(args.resources)
        self.run_singulairty = args.Singularity
        self.profiler_singularity_path = args.profiler_singularity_path
        self.snippy_singularity_path = args.snippy_singularity_path
        self.resistance_only = args.resistance_only
        self.detect_species = args.detect_species
        self.db_version = db_version
        self.jobs = args.jobs  
        self.profiler_threads = int(args.profiler_threads)
        self.kraken_threads = int(args.kraken_threads)
        self.kraken_db = args.kraken_db
        self.snippy_threads = int(args.snippy_threads)
        self.profiler_version = db_version
        self.isolates = ''
        # self.output = args.output
        self.set_snakemake_jobs()
        self.min_cov = args.min_cov
        self.min_aln = args.min_aln
        self.mode = args.mode 
        if self.mode == 'mdu':
            self.check_positive_control(args.positive_control)
            self.run_singulairty = True
        else:
            self.positive_control = {}
        self.check_rerun()


    def check_positive_control(self, positive_control):
        if positive_control == '' or not pathlib.Path(positive_control).exists():
            self.logger.warning(f"You are running Troika for MDU service. You must include a path positive control.")
            raise SystemExit
        else:
            p = pathlib.Path(positive_control)
            reads = sorted(p.glob(f"*.f*q.gz"))
            if len(reads) == 2:
                self.positive_control = {0:"2999-99887", 1:reads[0], 2:reads[1]}
                # return pos_string
            else:
                self.logger.warning(f"The path to positive control has not been found. Please check your settings and try again.")
                raise SystemExit
    
    def check_rerun(self):
        '''
        Check if this is a rerun of an existing job if so archive results into a dated dir
        '''
        report_toml = self.workdir / "report.toml"
        archive_dir = self.workdir / f"troika_output_{self.date}"
        self.logger.info(f"Checking if output files already exist")
        if report_toml.exists():
            self.logger.info(f"It seems you are running this in a pre-existing troika directory, existing files will be archived")
            if not archive_dir.exists():
                archive_dir.mkdir()
            ext_to_move = ["*.toml", "*.html", "*.csv", "new_isolates.txt"]
            for e in ext_to_move:
                cmd = f"mv {e} {archive_dir}"
                self.logger.info(f"{e} are/is being moved to {archive_dir}")
                subprocess.run(cmd, shell = True)
                


    def set_snakemake_jobs(self):
        '''
        set the number of jobs to run in parallel based on the number of cpus from args
        '''
        self.logger.info(f"Determining number of jobs to run in parallel.")
        threads = max([self.kraken_threads, self.profiler_threads, self.snippy_threads])
        # if int(self.jobs * threads) > int(psutil.cpu_count()):
        #     self.jobs = int(psutil.cpu_count()) - 1
    
        self.logger.info(f"Troika will run : {self.jobs} in parallel... keeping it nice.")
    
    def three_cols(self, tab):
        '''
        Ensure that there are 3 columns, isolate, R1 and R2
        returns True if 3 columns False otherwise
        
        '''
        print(tab)
        self.logger.info(f"Checking that input file is the correct structure. There are {tab.shape[1]} columns")
        if tab.shape[1] == 3:
            return True
        else:
            return False

    def all_data_filled(self, tab):
        '''
        Ensure that all fields contain data - no NA's
        returns True if there are no nan, False otherwise
        '''
        self.logger.info("Checking that there is no empty fields in the input file.")
        return tab.isnull().sum().sum() == 0
    
    def link_reads(self, read_source, isolate_id, r_pair):
        '''
        check if read source exists if so check if target exists - if not create isolate dir and link. If already exists report that a dupilcation may have occured in input and procedd

        '''
        # check that job directory exists
        self.logger.info(f"Checking that reads are present.")
        # check that READS exists
        R = self.workdir / isolate_id
        if not R.exists():
            R.mkdir()
        
        if f"{read_source}"[0] != '/':
            read_source = self.workdir / read_source
        
        if read_source.exists():
            read_target = R / f"{r_pair}"
            if not read_target.exists():
                read_target.symlink_to(read_source)
        else:
            self.logger.warning(f"{read_source} does not seem to a valid path. Please check your input and try again.")
            raise SystemExit()
    
    def path_exists(self,path, v = True):
        '''
        ensure files are present, if so continues if not quits with FileNotFoundError
        input:
            :path: patht to files for pipeline
            :v: if v == True print message, else just check
        output:
            returns True (or fails with FileNotFoundError)
        '''
        
        if not path.exists():
            self.logger.warning(f"The {path.name} does not exist.")
            raise FileNotFoundError(f"{path.name}")
        else:
            if v == True:
                self.logger.info(f"Found {path.name}.")

            return True

    def check_reads_exists(self, tab):
        '''
        check that the correct paths have been given
        if reads not present path_exists will cause a FileNotFound error and warn user
        :input
            :tab: dataframe of the input file
        '''
        self.logger.info(f"Checking that all the read files exist.")
        for i in tab.itertuples():
            
            if not '#' in i[1]:
                r1 = i[2]
                r2 = i[3]
                self.path_exists(pathlib.Path(r1), v = False)
                self.link_reads(pathlib.Path(r1), isolate_id=f"{i[1].strip()}", r_pair='R1.fq.gz')
                self.path_exists(pathlib.Path(r2), v = False)
                self.link_reads(pathlib.Path(r2), isolate_id=f"{i[1].strip()}", r_pair='R2.fq.gz')
        return True


    def get_samples(self):
        '''
        get samples list after ensuring that structure of input is correct and reads can be found
        '''

        tab = pandas.read_csv(f"{self.input_file}", sep = '\t', header = None)
        if not self.three_cols(tab):
            logging.warning(f"{self.input_file} does not appear to be in the correct configuration")
            raise TypeError(f"{self.input_file} has incorrect number of columns")
        
        if not self.all_data_filled(tab):
            self.logger.warning('warning',f"{self.input_file} appears to be missing some inforamtion.")
            raise TypeError(f"{self.input_file} appears to be missing some inforamtion.")
        
        if tab.shape[0] < 4:
            self.resistance_only = True
        
        if self.mode == 'mdu':
            dx = pandas.DataFrame(self.positive_control, index = [0])
            tab = tab.append(dx, sort = True)

        self.check_reads_exists(tab = tab)
        
        self.isolates = ' '.join([i for i in tab.iloc[:,0] if '#' not in i])
        self.logger.info(f"There are {len(list(tab.iloc[:,0]))} in this analysis")
    
    def check_input_file(self, path):
        '''
        Check that all files and directories exist
        '''
        self.logger.info(f"Searching for : {path}")
        if pathlib.Path(path).exists():
            self.logger.info(f"Success : {path} found.")
            return pathlib.Path(path)
        else:
            self.logger.warning(f"Failed : {path} was not found, please check your input and try again.")
            raise SystemExit
    
    def tbprofiler(self):
        '''
        check the installation of tb-profiler
        '''
        self.logger.info(f"Checking that tb-profiler is installed and recording version.")
        version_pat = re.compile(r'\bv?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<release>[0-9]+)(?:\.(?P<build>[0-9]+))?\b')
        try:
            tp = subprocess.run(['tb-profiler', 'version'], capture_output = True, encoding='utf-8')
            tp = tp.stdout
            # self.logger.info(f"{tp}")
            self.profiler_version = version_pat.search(tp).group(0)
            self.logger.info(f"TB-Profiler version {self.profiler_version} found. Good job!")
    
        except FileNotFoundError:
            self.logger.warning(f"TB-Profiler is not installed, please read installation/usage instructions and try again.")
            raise SystemExit

    def check_installation(self):
        '''
        If not using singularity then check for tbprofiler installation
        '''
        
        if not self.run_singulairty:
            self.tbprofiler()
        else:
            self.logger.info(f"You are using a singularity container today.")

    def generate_smk_config(self):
        '''
        Generate the config file
        '''
        config = {
                'template_path': f"{self.resources / 'templates'}",
                'script_path':f"{self.resources / 'utils'}", 
                'samples':self.isolates,
                'singularity_path_profiler' : self.profiler_singularity_path, 
                'singularity_path_snippy' : self.snippy_singularity_path, 
                'profiler_threads': self.profiler_threads, 
                # 'final_output' : final_output, 
                'db_version': self.db_version,
                'run_species': self.detect_species,
                'kraken_db': self.kraken_db,
                'kraken_threads':self.kraken_threads, 
                'snippy_threads': self.snippy_threads,
                'amr_only': self.resistance_only, 
                'reference': f"{self.resources / 'reference'/ 'tbdb.fasta'}", 
                'index': f"{self.resources / 'reference'/ 'tbdb.fasta.fai'}", 
                'mask': f"{self.resources / 'reference'/ 'mask.bed'}",
                'mode': self.mode,
                'positive_control': self.positive_control,
                'min_cov':self.min_cov,
                'min_aln':self.min_aln
        }

        config_source = self.resources / "templates" / "config.yaml"
        self.logger.info(f"Writing config file")
        config_template = jinja2.Template(config_source.read_text())
        config_target = self.workdir / "config.yaml"
        config_target.write_text(
            config_template.render(
                config
            )
        )
  
    def construct_command(self):
        '''
        once all checks have passed then construct the command
        '''
        # TODO add support for sbatch and qsub
        singularity = f"--use-singularity --singularity-args '--bind /home'" if self.run_singulairty else ''
        cmd = f"snakemake -s {self.resources / 'utils' / 'troika.smk'} -j {self.jobs} -d {self.workdir} {singularity}"
        return cmd
        

    def run_snakemake(self):
        '''
        run the snakemake command
        '''
        cmd = self.construct_command()
        self.logger.info(f"Now running AMR detection using the Troika pipeline with command {cmd}. Please be patient this may take some time.")
        wkf = subprocess.run(cmd, shell = True)
        while True:
            if wkf.stdout != None:
                line = wkf.stdout.readline().strip()
                if not line:
                    break
            line = ''
            break
            self.self.logger.info(f"{line}")
        if wkf.returncode == 0:
            return True
        else:
            return False


    def run_pipeline(self):
        '''
        run the pipeline
        '''
        
        # run checks
        self.logger.info(f"Checking TB-profiler installation.")
        self.check_installation()
        # check input file and get samples
        self.logger.info(f"Checking input file structure is correct and obtaining a list of samples.")
        self.get_samples()
        self.logger.info(f"Generating a config file for todays Troika job.")
        self.generate_smk_config()
        self.logger.info(f"All checks and preparations are completed.")
        if self.run_snakemake():
            self.logger.info(f"Checking that all outputs have been generated.")
            if self.mode == 'normal' and pathlib.Path(f"troika.tab").exists():
                self.logger.info(f"SUCCESS - all outputs have been generated. Have a nice day!")
            elif self.mode == 'mdu' and pathlib.Path(f"MMS155_{self.date}.csv").exists():
                self.logger.info(f"SUCCESS - all outputs have been generated. Have a nice day!")
            else:
                self.logger.warning(f"Something has gone wrong, output files have not been correcty created. Please check the logs and try again.")
    


import toml, pathlib, subprocess, sys, pandas, datetime

from snakemake import shell

def generate_iqtree_cmd(script_path, aln, reference):

    cmd = f"bash {script_path}/iqtree_generator.sh {reference} {aln} core 20"								
    return cmd

def generate_delete_cmd():
    cmd = f"rm *.ckp.gz *.contree *.bionj"
    return cmd

def run_cmd(cmd):

    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    print(p)
    return p.stdout

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs, ref, idx, script_path):
    
    t = open_toml(inputs)
    print(t)
    aln = 'core.aln'
    cmd = generate_iqtree_cmd(aln = aln, script_path = script_path, reference = ref)
    i = run_cmd(cmd)
    i = f"{i.strip()} -redo"
    print(i)
    iqtree = run_cmd(i)
    # print(iqtree)
    dl = generate_delete_cmd()
    rm = run_cmd(dl)
    data = {}
    data['iqtree'] = {}
    data['iqtree']['command'] = cmd
    data['iqtree']['input_file'] = aln
    data['iqtree']['file'] = 'core.treefile'

    write_toml(data = data, output = "iqtree.toml")
    
inputs = snakemake.input.core
ref = snakemake.input.ref
idx = snakemake.input.idx
script_path = snakemake.params.script_path

main(inputs = inputs, ref = ref, idx = idx,script_path = script_path)



import toml, pathlib, subprocess, sys, pandas, datetime, json

from snakemake import shell

def get_species(kraken):
    mtbc = ['pinnipedii', 'tuberculosis', 'orygis', 'bovis', 'bovis BCG', 'canetti', 'microti', 'africanum']
    tml = open_toml(kraken)
    # print(tml)
    isolate = list(tml.keys())[0]
    if tml[isolate]['kraken']['done'] == 'No':
        return 'kmer-id not performed'
    elif tml[isolate]['kraken']['done'] == 'Yes':
        s = tml[isolate]['kraken']['top_3'][0][0]
        # print(s)
        if 'Mycobacterium' in s:
            print('Mycobacterium genus found')
            for m in mtbc:
                # print(m)
                # print(m in s)
                if m in s:
                    print("found MTBC species")
                    return 'MTBC'
    # return 'non-MTBC'
            

def generate_tbprofiler_cmd(snippy, isolate, threads):
    tml = open_toml(snippy)
    bam = f"{tml[isolate]['snippy']['bam']}"
    cmd = f"tb-profiler profile -a {bam} --caller bcftools -t 6  -d {isolate}"

    return cmd

def generate_mv_cmd(isolate):

    cmd = f"mv {isolate}/results/tbprofiler.results.json {isolate}/tbprofiler.results.json"
    return cmd

def get_data(isolate):
    
    tb = {}
    if pathlib.Path(isolate,"tbprofiler.results.json").exists():
        d = json.load(open(f"{isolate}/tbprofiler.results.json"))
        tb['lineage'] = d['lineage']
        tb['main_lin'] = d['main_lin']
        tb['dr_variants'] = []
        if d['dr_variants'] != []:
            for dct in d['dr_variants']:
                list_keys = list(dct.keys())
                for k in list_keys:
                    if 'sample' in k:
                        dct.pop(k)
                tb['dr_variants'].append(dct)
        print(tb)
        # for i in d:
        #     print(i)
        #     print(d[i])
        return tb
    else:
        return {}

def generate_rm_cmd(isolate):

    cmd = f"rm -rf {isolate}/results {isolate}/bam {isolate}/vcf"
    return cmd

def run_cmd(cmd):

    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    # print(data)
    # for i in data:
    #         print(i)
    #         print(data[i])
    #         for j in data[i]:
    #             print(j)
    #             print(data[i][j])
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(snippy,qc, isolate, threads, kraken):
    
    # print(isolate)
    
    species = get_species(kraken)
    tml = open_toml(snippy)
    # print(tml)
    data = {}
    data[isolate] = {}
    data[isolate]['tb_profiler'] = {}
    print(data)
    print(tml)
    if tml[isolate]['snippy']['run_snippy'] == 'Yes' and species in ['MTBC', 'kmer-id not performed']:
        cmd = generate_tbprofiler_cmd(snippy = snippy, isolate = isolate, threads = threads)
        print(cmd)
        p = run_cmd(cmd)
        if p == 0:
            run_cmd(generate_mv_cmd(isolate = isolate))
            # run_cmd(generate_rm_cmd)
            data[isolate]['tb_profiler']['done'] = 'Yes'
            data[isolate]['tb_profiler']['kmer-id'] = species
            data[isolate]['tb_profiler']['data'] = get_data(isolate=isolate)
    else:
        data[isolate]['tb_profiler']['done'] = 'No'
        data[isolate]['tb_profiler']['kmer-id'] = 'kmer id not consistent with MTBC species'
        data[isolate]['tb_profiler']['data'] = get_data(isolate=isolate) 
    write_toml(data = data, output = f"{isolate}/tbprofiler.toml")   

snippy = snakemake.input.snippy
qc = snakemake.input.qc
kraken = snakemake.input.kraken
isolate = snakemake.wildcards.sample
threads = snakemake.params.threads
# snippy = f"2020-27522/snippy.toml"
# kraken = f"2020-27522/kraken.toml"
# isolate = f"2020-27522"
# threads = 1
# qc = f"2020-27522/snippy_qc.toml"
# 2020-27522
main(snippy = snippy, qc = qc, kraken = kraken, isolate = isolate, threads = threads)



import toml, pathlib, subprocess, sys, pandas

from snakemake import shell

def open_toml(tml):
    print(tml)
    data = toml.load(tml)
    print(data)
    return data
    
def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs):
    data = {}
    tab = pandas.DataFrame()
    for i in inputs:
        # print(i)
        d = open_toml(i)
        # print(list(d.keys()))
        isolate = list(d.keys())[0]
        df = pandas.DataFrame(data = d[isolate]['seqdata']['data'], index = [0])
        df['Isolate'] = isolate
        # print(df)
        if tab.empty:
            tab = df
        else:
            tab = tab.append(df)
    tab = tab.rename(columns = {'min_len': 'Min len', 'max_len': 'Max len', 'avg_len': 'Avg len', 'avgQ': 'Avg Qual', 'Estimated_coverage': 'Estimated depth', 'bases':'Yield'})
    tab = tab[['Isolate','Reads','Yield','GC content','Min len','Avg len','Max len','Avg Qual','Estimated depth', 'Quality']]
    tab.to_csv(f"seqdata.tab", sep = '\t', index = False)
    data['seqdata'] = tab.to_dict(orient = 'records')

    write_toml(data = data, output= "seqdata.toml")

inputs = snakemake.input

main(inputs = inputs)



# mash triangle -C *.msh

# mash sketch -m 5 -s 10000 -r -o 2019-12803-6/sketch -I 2019-12803-6 -C 2019-12803-6/R1.fq.gz 2019-12803-6/R1.fq.gz


import toml, pathlib, subprocess, sys, pandas, datetime

from snakemake import shell

def generate_dists_cmd(aln):
    
    cmd = f"snp-dists {aln} > distances.tab"								
    return cmd

def run_cmd(cmd):

    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.returncode

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs):
    
    t = open_toml(inputs)
    
    aln = 'core.aln'
    
    dists= run_cmd(generate_dists_cmd(aln = aln))
    if dists == 0:
        data = {}
        data['snp_dists'] = {}
        data['snp_dists']['input_file'] = aln
        data['snp_dists']['file'] = 'distances.tab'


        
        write_toml(data = data, output = "distances.toml")

inputs = snakemake.input
main(inputs = inputs)



import json,pathlib,pandas, sys, datetime, toml, numpy

from snakemake import shell

def return_species(data):
    '''
    data is the dictionary from json.load
    '''
    
    if data['snpit']['species'] != 'N/A':
        return data['snpit']['species']
    elif 'BOV' not in data['lineage'][0]['lin']:
        return 'M. tuberculosis'
    elif 'BOV' in data['lineage'][0]['lin']:
        return 'non-tuberculosis MTBC'
    else:
        return "species undefined"
    

def return_lineage(data):
    '''
    data is the dictionary from json.load
    '''

    if data['lineage'] != []:
        main_lineage = data['main_lin'].replace('lineage', 'Lineage ')
        semicolon_check = main_lineage.split(';')
        if len(semicolon_check) > 1 :
            lins = [s.split('.')[0] for s in semicolon_check]
            if len(set(lins)) > 1:
                lineage_string = 'phylogenetic lineage undefined' 
            else:
                f = data['lineage'][0]['family'].split()[0]
                lineage_string = f"{f} ({lins[0]})"
        else:
            lineage_string = f"{data['lineage'][0]['family']} ({main_lineage})"
    elif data['lineage'] == [] and data['snpit']['lineage'] != 'N/A':
        lineage_string = f"{data['snpit']['family']} ({data['snpit']['lineage']})"
    else:
        lineage_string = 'phylogenetic lineage undefined' 

    return lineage_string

def construct_string(mutation):

    gene = mutation['gene_name'] if 'gene_name' in mutation else mutation['gene']
    change = mutation['change']
    gene = f"{gene[0].upper()}{gene[1:]}"
    return f"{gene} ({change})"    

def check_variants(mutation, drug):
    '''
    mutation is the dictionary from dr_variants
    based on validation of mms155
    allelic frequency minimum set to .1 so no need to filter
    for Rif, Inh and Emb - use mutations that are NOT indeterminate
    for all other drugs - use ALL mutations in the dr_variants - ie include the indeterminate
    '''
    
    exclude_ind = ['rifampicin', 'isoniazid', 'ethambutol']
    if drug in exclude_ind and mutation['confidence'] != 'indeterminate':
        return construct_string(mutation = mutation)
    elif drug not in exclude_ind:
        return construct_string(mutation = mutation)
    else:
        return ''

def get_dr_variants(data):
    '''
    data is the dictionary from json.load
    if drug in 'rifampicin', 'isoniazid', 'ethambutol' the EXCLUDE 'indeterminate' mutations, eles include them 
    '''
    # Check with Norelle - whether or not she wants to include drugs which did not fall in the validated dataset
    # a dictionary defaulted to 'No mutation detected'
    d = {
        'Amikacin':'No mutation detected', #Block 2
        # 'Aminoglycosides':'No mutation detected',
        'Bedaquiline':'No mutation detected',
        'Capreomycin':'No mutation detected',
        'Ciprofloxacin':'No mutation detected', 
        'Clofazimine':'No mutation detected',
        'Cycloserine':'No mutation detected',
        'Delamanid':'No mutation detected', #Block 3
        'Ethambutol':'No mutation detected', 
        'Ethionamide':'No mutation detected', 
        'Fluoroquinolones':'No mutation detected', 
        'Isoniazid':'No mutation detected',
        'Kanamycin':'No mutation detected', 
        'Levofloxacin':'No mutation detected', 
        'Linezolid':'No mutation detected', 
        'Moxifloxacin':'No mutation detected', #Block 1
        'Ofloxacin':'No mutation detected', 
        'Para-aminosalicylic_acid':'No mutation detected',
        'Pyrazinamide':'No mutation detected',
        'Rifampicin':'No mutation detected',
        'Streptomycin': 'No mutation detected'}
    
    
    if data['dr_variants'] != []:
        for mut in data['dr_variants']:
            drug = mut['drug']
            mutation = check_variants(mutation = mut, drug = drug)
            if mutation != '' and drug not in ['aminoglycosides']: # TODO streptomycin needs to be added back in
                if d[f'{drug.capitalize()}'] != 'No mutation detected':
                    d[f'{drug.capitalize()}'] = f"{ d[f'{drug.capitalize()}']},{mutation}"
                else:
                    d[f'{drug.capitalize()}'] = mutation
    
    return d

def calculate_resistance(drugs_dict):
    '''
    based on WHO criteria for TB resistance categories
    No drug resistance predicted = no mutations in first line drugs
    Mono-resistance predicted =resistance in one of rif, inh, emb or pza
    Poly-resistance predicted  = two or more resisance where both rif and inh are not included
    Multi-drug resistance predicted = rif and inh resistance plus or minus emb and pza
    Extensive drug-resistance predicted rif and inh AND a flouroquinolone AND one of amikacin/capreomycin/kanamycin/sm #TODO add Streptomycin to logic
    '''
    # list of drigs in flq or ack for xdr
    flqlist = ['Fluoroquinolones','Levofloxacin','Moxifloxacin','Ofloxacin','Ciprofloxacin']
    ackslist = ['Amikacin', 'Capreomycin', 'Kanamycin', 'Streptomycin']
    # initialise score 
    score = 0
    # initialise flq and ack values
    flq = False
    ack = False
    # generate values for resistance call
    for d in drugs_dict: 
        if drugs_dict[d] != 'No mutation detected':
            if d in flqlist:
                flq = True
            elif d in ackslist:
                ack = True
            elif d in ['Rifampicin', 'Isoniazid']:
                score = score + 3
            elif d in ['Pyrazinamide', 'Ethambutol']:
                score = score + 1
    
    # initialise the resistance to default 
    resistance = 'No drug resistance predicted'
    if score == 1 or score == 3: # one first line drug
        resistance = 'Mono-resistance predicted'
    elif score == 2 or score in range(4,6): # more than one first line drug where INH OR RIF can be present
        resistance = 'Poly-resistance predicted'
    elif score in range(6,9): # RIF and INH +/- PZA or EMB
        if flq and ack: # there are mutations in a FLQ and amikacin,capreomycin,kanamycin
            resistance = 'Extensive drug-resistance predicted'
        else:
            resistance = 'Multi-drug resistance predicted'
    
    return resistance

def make_df(result_dict):

    cols_list = ['Isolate', 'Organism identification by WGS', 'Phylogenetic lineage', 'Predicted drug resistance','Rifampicin', 'Isoniazid', 'Pyrazinamide', 'Ethambutol','Moxifloxacin','Amikacin', 'Cycloserine', 'Ethionamide', 'Para-aminosalicylic acid','Clofazimine', 'Delamanid', 'Bedaquiline', 'Linezolid','Database version used for analysis']
    
    df = pandas.DataFrame.from_dict(data = result_dict, orient = 'index')
    df = df.reset_index()
    df = df.rename(columns = {'index': 'Isolate', 'Para-aminosalicylic_acid':'Para-aminosalicylic acid'})
    df = df.reindex(cols_list, axis = 'columns')
    df = df.fillna('No mutation detected')
    # print(df)
    return df
# ERR2120246
def save_results(result_dict):

    write_toml(data = result_dict, output = 'resistance.toml')
    df = make_df(result_dict)
    df.to_csv(f'troika.tab', index = False, sep = '\t')


def mdu_troika(result_dict):
    date = datetime.datetime.today().strftime("%d_%m_%y")
    edit_list = ['Phylogenetic lineage', 'Predicted drug resistance','Rifampicin', 'Isoniazid', 'Pyrazinamide', 'Ethambutol','Moxifloxacin','Amikacin', 'Cycloserine', 'Ethionamide', 'Para-aminosalicylic acid','Clofazimine', 'Delamanid', 'Bedaquiline', 'Linezolid']
    df = make_df(result_dict)
    for e in edit_list:
        df[e] = numpy.where(df['Organism identification by WGS'] != 'M. tuberculosis', '', df[e])
    df = df.rename(columns = {'Isolate': 'MDU sample ID',
                               'Organism identification by WGS': 'Identification (WGS)',
                               'Predicted drug resistance':'Predicted drug resist. summary:'})
    df = df.replace('No mutation detected','No mutn det')
    df.to_csv(f'MMS155_{date}.csv', index = False)

def collate_results(isolates, db, mode):
    
    # iszolates = isolates.split(',')
    result_dict = {}
    for isolate in isolates:
        p = pathlib.Path(isolate)

        json_path = p / f"tbprofiler.snpit_results.json"
        print(json_path)
        data = json.load(open(json_path))
        print(data)
        species = return_species(data)
        print(species)
        lineage = return_lineage(data)
        print(lineage)
        drugs = get_dr_variants(data)
        print(drugs)
        resistance = calculate_resistance(drugs)
        print(resistance)
        
        d = {
            'Organism identification by WGS':species,
            'Phylogenetic lineage':lineage,
            'Database version used for analysis':db,
            'Predicted drug resistance':resistance
            }
        for drug in drugs:
            d[drug] = drugs[drug]
        result_dict[isolate] = d
    # output_name = output[0].split('.')[0]
    save_results(result_dict)
    if mode == 'mdu':
        mdu_troika(result_dict)

    # if mdu:
    #     results_mdu = mdu_troika(result_dict)



# results to be recorded by tbrnr
# Species -> from snpit M. tuberculosis
# Lineage -> from TB-profiler if species == M. tuberculosis, if no lineage detected with TB-profiler then revert to snpit
# dr_mutations -> Drug -> comma separated list of mutations
# Drug resistance summary -> based on WHO criteria
# csv columns <MDU ID>,<WGS species>,<Lineage>,<Predicted drug resistance>,<Rif (or other drug names)>,<DB version>
# json 
# {
#     MDU ID : {
#         WGS species,
#         Lineage,
#         Predicted drug resistance,
#         Rif or other drug,
#         DB version
#     }
# }

def open_toml(tml):

    data = toml.load(tml)

    return data

def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)

def main(db, inputs, mode):
    
    isolates = []
    for i in inputs:
        tml = open_toml(i)
        isolate = list(tml.keys())[0]
        # print(list(tml.keys())[0])
        # print(tml)
        if tml[isolate]['snpit']['done'] == 'Yes' and tml[isolate]['tb_profiler']['done'] == 'Yes':
            isolates.append(isolate)
    collate_results(isolates=isolates, db=db, mode = mode)
    
db = snakemake.params.db_version
mode = snakemake.params.mode
inputs = snakemake.input

main(db = db,mode = mode, inputs = inputs)


import toml, pathlib, subprocess, sys, datetime, pandas, re, numpy


def write_tables(self,reportdir, table):
        '''
        Write a table, given a tab delimited file generate a html string
        '''
        # TODO add class isolate id to <tr>
        # TODO add class distances-isolate to tr if matrix and head-isolate to head td
        path = f"{table}"
        if path.exists():
            data = open(path).readlines()
            # for header
            header = data[0].split('\t') 
            if 'distances.tab' in table:
                tablehead = [f"<th class='{column}-head'>{column}</th>" for column in header]
            else:
                tablehead = [f"<th>{column}</th>" for column in header]
            # for body seqtablebody
            body = []
            for i in range(1,len(data)):
                raw = data[i].split('\t')  
                if 'summary_table.tab' in table:
                    row = [f"<tr class='{raw[0]} tiplab'>"]
                elif 'distances.tab' in table:
                    row = [f"<tr class='distances-{raw[0]}'>"]
                elif 'mtb' in table:
                    row = [f"<tr class='{raw[0]}-species-identification'>"]
                elif 'core_genome.tab' in table:
                    row = [f"<tr class='{raw[0]}-core-genome'>"]
                elif 'troika.tab' in table:
                    row = [f"<tr class='{raw[0]}-drug-resistance'>"]
                elif 'seqdata.tab' in table:
                    row = [f"<tr class='{raw[0]}-sequence-data'>"]
                else:
                    row = [f"<tr>"] # TODO add class isolate id to <tr>
                if 'distances.tab' in table:
                    for d in range(len(raw)):
                        row.append(f"<td align=\"center\" class = \"{raw[0]}_{header[d]}\">{raw[d]}</td>")    
                else:
                    for d in raw:
                        row.append(f"<td align=\"center\">{d}</td>")
                row.append(f"</tr>")
                body = body + row
            return('\n'.join(tablehead),'\n'.join(body))

def get_isolates_preview(table):
    
    data = open(table).readlines()
    isolates = []
    for d in range(1,len(data)):
        l = data[d].split()
        isolates.append(l[0])
    return len(isolates)

def preview_distances_tab(table):
    
    data = open(table).readlines()
    length = get_isolates_preview(table)
    header = ['mash-dist']
    
    body = []
    for d in range(1,len(data)):
        row = [f"<tr>"]
        l = data[d].split()
        print(l)
        header.append(l[0])
        for i in range(len(l)):
        # print(d)  
            row.append(f"<td align=\"center\" class = \"{l[0]}_{header[i]}\">{l[i]}</td>")
        row.append(f"</tr>")
        body = body + row
    # print(body)
    tablehead = [f"<th class='{column}-head'>{column}</th>" for column in header]
    return('\n'.join(tablehead),'\n'.join(body))

def get_tree_string(pipeline):
    '''
    Generate a tree image from a newick
    input:
        :reportdir: the directory where report files are stored
    output:
        string reporesentation of the path to the tree image
    '''
    tree_file = 'preview.newick' if pipeline == 'preview' else 'core.treefile'
    with open(f"{tree_file}", 'r') as t:
        tree = t.read().strip()

    return tree
def adjust_offset(row, d):
    if row['CHR'] in d:
        return(int(d[row['CHR']]) + int(row['POS']))
    else:
        return(int(row['POS']))

def generate_dict(idx_file):
    d = {}
    pos = 0
    offset = 0
    with open(f"{idx_file}") as f:
        for line in f:
            l = line.split()
            d[l[0]] = offset
            offset = offset + int(l[1])
    return(d)

def plot_snpdensity():

    '''
    generate a snp-density accross the genome plot - using the core.tab file
    input:
        :reportdir: the directory where files are kept
    out:
        :snpdensityscript: the javascript string for insert into html
        :snpdenstiydiv: the html div for inster in html doc
    '''
    # helper functions for getting the data into the right format.
    
    
    # open fai file and generate the dictionary
    idx = pathlib.Path('ref.fa.fai')
    d = generate_dict(idx)
    # get the core file
    core = 'core.tab'
    df = pandas.read_csv(core, sep = '\t')
    # get a list of isolate names
    names = list(df.columns[3:len(df.columns)])
    # if the there is no snp in the isolate (ie same as ref then mak na - then easy to drop)
    for i in names:
        df[i]=numpy.where(df['REF'] == df[i], numpy.nan, df[i])
    # generate the offset value
    df['POS_OFFSET'] = df[['CHR', 'POS']].apply(lambda x:adjust_offset(x,d), axis = 1)
    # collect positions to get allow for histogram and dropna (no snp)
    melted_df = pandas.melt(df, id_vars=['POS_OFFSET'], value_vars=names)
    melted_df = melted_df.dropna()
    melted_df = melted_df.sort_values(by= ['POS_OFFSET'])
    # generate histogram
    # snpdensityscript, snpdensitydiv = self.plot_histogram(series=melted_df['POS']/1000,xlabel="Genome Position (MB)", ylabel="SNPS",bins=10000)
    contig_breaks = [d[value] for value in d]
    
    # return dictionary
    return(list(melted_df['POS_OFFSET']/1000))


def plot_distances():

    '''
    generate a snp-density plot - using the distacnes.tab file
    input:
        :reportdir: the directory where files are kept
    out:
        :distancescript: the javascript string for insert into html
        :distancesdiv: the html div for inster in html doc
    '''
    distance = 'distances.tab'

    df = pandas.read_csv(distance, sep = '\t')
    # get a list of isolate names
    names = list(df.columns[1:len(df.columns)])
    col1 = df.columns[0]
    
    # if the there is no snp in the isolate (ie same as ref then mak na - then easy to drop)
    
    # collect positions to get allow for histogram and dropna (no snp)
    melted_df = pandas.melt(df, id_vars=[col1], value_vars=names)
    melted_df = melted_df[melted_df[col1]!= melted_df['variable']]
    # generate histogram
    # distancescript, distancediv = self.plot_histogram(series=melted_df['value'], xlabel="SNP distances", ylabel="Frequency", bins=100)
    # td['pairwisedistance'] = {'script':distancecript, 'div':distancediv}
    # return dictionary
    return(list(melted_df['value']))

def get_preview_dict():

    td = [{'title': 'Preview', 'link':'preview-tree', 'file': 'preview.newick', 'type': 'tree'},
               {'file': 'preview_distances.tab', 'title':'Mash distances', 'type':'matrix', 'link':'mash-distances'}
        ]
    return td

def snps_dict(td):
    
    core_genome_td = {'file': 'core_genome.tab', 'title': 'Core Genome', 'link':'core-genome', 'type':'table'}
    snp_density_td = {'title': 'SNP density', 'link':'snp-density', 'type':'graph'}
    core_phylogeny_td = {'title': 'Phylogeny', 'link':'phylogeny', 'file': 'core.treefile', 'type': 'tree'}
    snp_distance_td = {'file': 'distances.tab', 'title':'SNP distances', 'type':'matrix', 'link':'snp-distances'}
    snps = [core_genome_td, snp_density_td, core_phylogeny_td,snp_distance_td]
    for s in snps:
        td.append(s)
    
    return td

def assembly_dict(td):
    for_td = []
    mlst_td = {'file':'mlst.tab', 'title':'MLST', 'type':'table', 'link':'mlst'}
    resistome_td = {'file':'resistome.tab', 'title':'Resistome', 'type':'table', 'link':'resistome'}
    # list of assembly tasks
    assembly_stat_td = {'file': 'assembly.tab', 'title':'Assembly', 'type':'table', 'link':'assembly'}
    a_td = [assembly_stat_td, mlst_td, resistome_td]
    for a in a_td:
        if pathlib.Path(a['file']).exists():
            for_td.append(a)
    for f in for_td:
        td.append(f)

    return td

def sa_dict(td):

    td = snps_dict(td)
    td = assembly_dict(td)

    return td

def roary_dict(td):

    roary_td = {'file':'roary/summary_statistics.txt', 'title':'Pan Genome', 'type': 'pan', 'image': f"{pathlib.Path('pan_genome.svg').open().read()}", 'link':'pan-genome'}

    td.append(roary_td)

    return td

def all_dict(td):

    td = sa_dict(td)
    td = roary_dict(td)

    return td

def get_dict(pipeline):

    td = [{'file':'seqdata.tab', 'title':'Sequence Data', 'link': 'sequence-data', 'type' : 'table'}, {'file':'species_identification.tab', 'title': 'Species Identification', 'type': 'table', 'link':'species-identification'},{'file':'summary_table.tab','title':'Summary', 'link':'summary', 'type':'summary'},{'file': 'software_versions.tab', 'title': 'Tools', 'type': 'versions', 'link':'versions'}]
    
    
    if pipeline == 's':
        td = snps_dict(td)
    elif pipeline == 'a':
        td = assembly_dict(td)
    elif pipeline == 'sa':
        td = sa_dict(td)
    elif pipeline == 'all':
        td = all_dict(td)
    # print(td)
    return td

def fill_vals(td, pipeline):

    # if pipeline == 'preview':
    #     for t in range(len(td)):
    #         if td[t]['type'] == 'matrix':
    #             td[t]['head'], td[t]['body'] = preview_distances_tab(table=td[t]['file'])
    
    # # snpdistances = []
    # # snpdensity = []
    # else:
    print(range(len(td)))
    # print(td[10])
    for t in range(len(td)):
        # print(t)
        # print(td[t])

        # TODO if table add a modal modal + link and link will be title lowercase with hyphen
        if td[t]['type'] == 'table':
            td[t]['head'], td[t]['body'] = write_tables(table=td[t]['file'])
        if td[t]['type'] == 'pan':
            td[t]['head'], td[t]['body'] = write_tables(table=td[t]['file'])
        if td[t]['type'] == 'matrix':
            td[t]['head'], td[t]['body'] = write_tables(table=td[t]['file'])
            # snpdistances = plot_distances()
        # if td[t]['link'] == 'snp-density':
            # snpdensity= plot_snpdensity()
        if td[t]['type'] == 'versions':
            td[t]['head'], td[t]['body'] = write_tables(table=td[t]['file'])
        if td[t]['type'] == 'summary':
            # generate_summary()
            td[t]['head'], td[t]['body'] = write_tables(table = td[t]['file'])
        print(t)
    return td

def get_software_versions(software):

    '''
    Given the name of the software, find the version
    input:
        :software: the name of the software
    output:
        a string in the form of 'Name_of_Sofware v.X.Y.Z'
    '''
    
    version_pat = re.compile(r'\bv?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<release>[0-9]+)(?:\.(?P<build>[0-9]+))?\b')

    if software == 'snp-dists':
        vs = '-v'
    else:
        vs = '--version'
    cmd = f"{software} {vs} 2>&1"
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = "utf-8")
    sft = p.stdout
    
    v = version_pat.search(sft)
    v = v.group()
    sft_version = f"{software} v.{v}"
    return(sft_version)

def make_dict_versions(tools):
    '''
    Called by get_software_file to make a dictionary of tools used
    input:
        :tools: a list of tools
    output:
        a dictionary with tools and versions.
    '''
    tool_dict = {}
    for t in tools:
            v = get_software_versions(t)
            tool_dict[t] = v
    return(tool_dict)


def get_software_file(pipeline, assembler = ''):
    '''
    get the versions of software on the system at completion of the piepline
    input:
        :reportdir: the directory where report files are stored
        :pipeline: the type of pipeline
        :assembler: the assembler used in the pipeline
    '''
    snippy_tools = ['snippy', 'snippy-core', 'snp-dists', 'iqtree']
    assembly_tools = ['mlst', 'kraken2', 'prokka', 'abricate', assembler]
    
    if pipeline == 's':
        tool_dict = make_dict_versions(snippy_tools)
    elif pipeline == 'a':
        tool_dict = make_dict_versions(assembly_tools)
    elif pipeline == 'sa':
        snippy_tools.extend(assembly_tools)
        tool_dict = make_dict_versions(snippy_tools)
    elif pipeline == 'all':
        snippy_tools.extend(assembly_tools)
        snippy_tools.append('roary')
        tool_dict = make_dict_versions(snippy_tools)
    
    
    versions = ['Software versions']
    for t in tool_dict:
        versions.append(tool_dict[t])
    
    p = pathlib.Path('software_versions.tab')

    p.write_text('\n'.join(versions))

def merge_dfs(start, added):
    if start.empty:
        start = added
    else:
        start = start.merge(added, how = 'outer')
    return(start)

def generate_summary():
    '''
    function to generate a summary table
    '''
    p = pathlib.Path('.')
    tabs = [t for t in p.iterdir() if f"{t.suffix}" == '.tab']
    # print(tabs)
    summary_df = pandas.DataFrame()
    df_list = []
    
    # print(tabs)
    for tab in tabs:
        # print(df)
        # print(tab)
        if 'species' in f"{tab.name}":
            species = pandas.read_csv(tab, sep = '\t')
            species = species[['Isolate', 'Match #1']]
            summary_df = merge_dfs(summary_df, species)
            summary_df = summary_df.rename(columns={'Match #1': 'Species'})
        elif 'seqdata' in f"{tab.name}":
            seq = pandas.read_csv(tab, sep = '\t')
            seq = seq[['Isolate', 'Estimated depth']]
            summary_df = merge_dfs(summary_df, seq)
        elif 'assembly' in f"{tab.name}":
            assembly = pandas.read_csv(tab, sep = '\t')
            assembly = assembly[['Isolate', '# Contigs']]
            summary_df = merge_dfs(summary_df, assembly)
        elif 'mlst' in f"{tab.name}":
            mlst = pandas.read_csv(tab, sep = '\t', skiprows=1, header=None)
            mlst = mlst.rename(columns = {0:'Isolate', 2:'ST'})
            mlst = mlst[['Isolate', 'ST']]
            summary_df = merge_dfs(summary_df, mlst)
        elif 'core_genome' in f"{tab.name}":
            core = pandas.read_csv(tab, sep = '\t')
            # print(core)
            core = core[['Isolate', '% USED']]
            summary_df = merge_dfs(summary_df, core)
    isolates = len(summary_df['Isolate'])
    # print(mlst)
    # print(summary_df)
    summary_df = summary_df.fillna('NA')
    summary_file = 'summary_table.tab'
    summary_df.to_csv(summary_file, sep = '\t', index = False)

    return isolates

def return_tables(pipeline):

    if pipeline == 's':
        
        tables =['core-genome', 'snp-distances', 'sequence-data','species-identification']
        modaltables =['core-genome',  'sequence-data','species-identification']
        display = f"display:inline;"
    elif pipeline == 'a':
        
        tables =['mlst', 'assembly', 'resistome', 'sequence-data','species-identification']
        modaltables = ['mlst', 'assembly', 'resistome', 'sequence-data','species-identification']
        display = f"display:none;"
    elif pipeline == 'sa':
        
        tables =['core-genome', 'snp-distances', 'mlst', 'assembly', 'resistome', 'sequence-data','species-identification']
        modaltables = ['core-genome',  'mlst', 'assembly', 'resistome', 'sequence-data','species-identification']
        display = f""
        # td.extend(s_td)
    elif pipeline == 'all':
        # a_ll = td.extend()
        tables =['core-genome', 'species-identification','snp-distances', 'mlst', 'assembly', 'resistome', 'sequence-data', 'pan-genome']
        modaltables = ['core-genome',  'mlst', 'assembly', 'resistome', 'sequence-data']
        display = f""
    
    tables.append('versions')

    return tables, modaltables, display

def open_toml(tml):

    data = toml.load(tml)

    return data
    
def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(inputs, pipeline,job_id, assembler = ''):
    # initialise dictionary
    # print(pipeline)
    print(inputs)
    data = {
        'newick' :'',
        'snpdensity':'',
        'snpdistances':'',
        'display':'',
        'job_id':job_id,
        'pipeline':pipeline,
        'date':datetime.datetime.today().strftime("%d_%m_%y"), 
        'tree_heigth':0,
        'modaltables':'',
        'tables':''
        }
    # print(data)
    if pipeline == 'preview':
        td = get_preview_dict()
        data['tree_height'] = get_isolates_preview('preview_distances.tab') * 25
        data['tables'] = ['mash-distances']
        data['modaltables'] = ['mash-distances']
        # print(td)
    else:
        isos = generate_summary()
        # print(isos)
        get_software_file(pipeline = pipeline, assembler = assembler)  
        td = get_dict(pipeline = pipeline)
        tables, modaltables, display = return_tables(pipeline = pipeline)
        data['tables'] = tables
        data['modaltables'] = modaltables
        data['display'] = display
        data['tree_heigth'] = isos * 25

    if pipeline not in ['a', 'preview']:
        data['snpdensity']= plot_snpdensity()
        data['snpdistances']= plot_distances()
        data['newick'] = get_tree_string(pipeline = pipeline)
    elif pipeline in ['s', 'sa', 'preview', 'all']:
        data['newick'] = get_tree_string(pipeline = pipeline)
    

# newick = newick, display = display,tables = tables,td = td, job_id = job_id, pipeline = pipeline, snpdistances=snpdistances, snpdensity = snpdensity, modaltables = modaltables, date = date
    td = fill_vals(td=td, pipeline = pipeline)
    data['td'] = td
    # print(data)
    write_toml(data = data, output = 'report.toml')    


if __name__ == '__main__':
    
    main(inputs = sys.argv[4:], pipeline = f"{sys.argv[1]}", job_id = f"{sys.argv[2]}",assembler = f"{sys.argv[3]}")
    



# mash triangle -C *.msh

# mash sketch -m 5 -s 10000 -r -o 2019-12803-6/sketch -I 2019-12803-6 -C 2019-12803-6/R1.fq.gz 2019-12803-6/R1.fq.gz


import toml, pathlib, subprocess, sys
from snakemake import shell

def generate_cmd(r1, r2, isolate):
    
    cmd = f"mash sketch -r {r1} -m 5 -k 21 -C {isolate} -o {isolate}/sketch"
    return cmd

def run_cmd(cmd):
    
    p = subprocess.run(cmd, shell = True, capture_output=True, encoding = 'utf-8')
    return p.stderr

def extract_metrics(mash_string):
    mash_string = mash_string.split('\n')
    d = ''
    for m in mash_string:
        print(m)
        if 'Estimated coverage' in m:
            d = m.split(':')[-1].strip()
            print(d)
            return d
    
def write_toml(data, output):
    
    with open(output, 'wt') as f:
        toml.dump(data, f)
    
def main(r1, r2, isolate, output):
    
    cmd = generate_cmd(r1 = r1, r2 = r2, isolate = isolate)
    mash_string = run_cmd(cmd)
    # print(mash_string)
    data = {}
    data[isolate] = {}
    data[isolate]['mash'] = {}
    data[isolate]['mash']['Estimated coverage'] = extract_metrics(mash_string)
    data[isolate]['mash']['sketch'] = f"{pathlib.Path(f'{isolate}', 'sketch.msh')}"

    write_toml(data = data, output= f'{isolate}/mash.toml')
 

r1 = snakemake.input.r1
r2 = snakemake.input.r2
isolate = snakemake.wildcards.sample
output = snakemake.output

main(r1 = r1, r2 = r2, isolate = isolate, output = output)


import sys, pathlib, pandas, pytest, numpy

from unittest.mock import patch

from troika.RunTroika import RunTroika



def test_3_cols():
        '''
        return True when correct number of columns
        '''
        with patch.object(RunTroika, "__init__", lambda x: None):
                detect_obj = RunTroika()
                tab = pandas.DataFrame({'A':[1], 'B':[2], 'C':[3]})
                assert detect_obj.three_cols(tab)


def test_2col_dimensions():
        '''
        return False when wrong number of columns present
        '''
        with patch.object(RunTroika, "__init__", lambda x: None):
                detect_obj = RunTroika()
                tab = pandas.DataFrame({'A':[1], 'B':[2]})
                assert detect_obj.three_cols(tab) == False


def test_path_exists():
        '''
        test that path_exists returns True
        '''
        with patch.object(RunTroika, "__init__", lambda x: None):
                p = pathlib.Path('troika','tests', 'troika.txt')
                detect_obj = RunTroika()
                assert detect_obj.path_exists(p)




import pandas, re, pathlib, sys




def get_data(output):
#    print(seqtkdata[0]
    summary_pat = re.compile(r'(\w+):\s(\d+\.?\d?\d?);')
    distinct_values = re.compile(r';\s(\d+)\s\w+')
    # print(distinct_values)
    print(output[1])
    record = re.compile(r'(\d+\.?\d?\d?)')
    summary_dict = dict((k, float(v)) for k, v in summary_pat.findall(output[0]))
    # print(distinct_values.findall(output[0]))
    summary_dict['distinct_error_codes'] = int(distinct_values.findall(output[0])[0])
    header = output[1].replace('%', '').replace('#', '').replace("POS\t", "").split("\t")
    line_one = [float(v) for v in record.findall(output[2])]
    summary_dict.update(dict(zip(header, line_one)))
    header.insert(0, 'position')
    summary_dict['Reads'] = dict(zip(header, [float(v) for v in record.findall(output[3])]))['bases']
    summary_dict['GC content'] = round((summary_dict['C'] + summary_dict['G']),2)
    median_position = (summary_dict['Reads'] + 1) / 2
    for line in output[4:]:
        parsed_record = dict(zip(header, [float(v) for v in record.findall(line)]))
        if parsed_record['bases'] > median_position:
            continue
        else:
            summary_dict['med_len'] = parsed_record['position'] - 1
            break
    return(summary_dict)

def get_coverage(mashdata):

    return(float(mashdata[1].strip().split(':')[1].strip()))


def main(pathtoseqtkdata, pathtomashdata, outputpath):
    output = pathlib.Path(f"{pathtoseqtkdata}").open().readlines()
    print(output)
    seqtkdata = get_data(output)
    df = pandas.DataFrame(data = seqtkdata, index = [0])
    df = df[['Reads','bases','GC content','min_len', 'avg_len', 'max_len','avgQ']]
    mash = pathlib.Path(f"{pathtomashdata}").open().readlines()
    df['Estimated depth'] = int(get_coverage(mash))
    df = df.rename(columns={'bases':'Yield', 'min_len': 'Min len', 'avg_len': 'Avg len', 'max_len':'Max len', 'avgQ': 'Avg Qual'})
    print(df.head())
    df = df[['Reads','Yield','GC content','Min len','Avg len','Max len','Avg Qual','Estimated depth']]
    df.to_csv(pathlib.Path(f"{outputpath}"), index = False, sep = '\t')

  

if __name__ == '__main__':
    
    # print(sys.argv[1], sys.argv[2], sys.argv[3])
    main(f"{sys.argv[1]}", f"{sys.argv[2]}", f"{sys.argv[3]}")